# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import factory
from yafowil.base import fetch_value
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_positional_rendering_helper
from yafowil.common import generic_required_extractor
from yafowil.compat import ITER_TYPES
from yafowil.compat import STR_TYPE
from yafowil.datatypes import convert_value_to_datatype
from yafowil.datatypes import convert_values_to_datatype
from yafowil.datatypes import generic_datatype_extractor
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.utils import EMPTY_VALUE
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops
from yafowil.utils import vocabulary


###############################################################################
# select
###############################################################################

@managedprops('multivalued', 'disabled')
def select_extractor(widget, data):
    extracted = generic_extractor(widget, data)
    multivalued = attr_value('multivalued', widget, data)
    exists_marker = '{0}-exists'.format(widget.dottedpath)
    if extracted is UNSET and exists_marker in data.request:
        if multivalued:
            extracted = []
        else:
            extracted = ''
    if extracted is UNSET:
        return extracted
    if multivalued and isinstance(extracted, STR_TYPE):
        extracted = [extracted]
    disabled = widget.attrs.get('disabled', False)
    if not disabled:
        return extracted
    if not multivalued:
        return data.value
    disabled_items = disabled is True and data.value or disabled
    if isinstance(disabled_items, STR_TYPE):
        disabled_items = [disabled_items]
    for item in disabled_items:
        if item in extracted and item not in data.value:
            extracted.remove(item)
        elif item not in extracted and item in data.value:
            extracted.append(item)
    return extracted


def select_exists_marker(widget, data):
    tag = data.tag
    attrs = {
        'type': 'hidden',
        'value': 'exists',
        'name_': "{0}-exists".format(widget.dottedpath),
        'id': cssid(widget, 'exists'),
    }
    return tag('input', **attrs)


def select_edit_renderer_props(widget, data):
    value = fetch_value(widget, data)
    multivalued = attr_value('multivalued', widget, data)
    if isinstance(value, STR_TYPE) or not hasattr(value, '__iter__'):
        value = [value]
    datatype = widget.attrs.get('datatype', None)
    if datatype:
        value = convert_values_to_datatype(value, datatype)
    emptyvalue = attr_value('emptyvalue', widget, data, EMPTY_VALUE)
    if not multivalued and len(value) > 1:
        raise ValueError(u'Multiple values for single selection.')
    disabled = attr_value('disabled', widget, data, False)
    return value, multivalued, datatype, emptyvalue, disabled


def select_block_edit_renderer(widget, data, custom_attrs={}):
    value, multivalued, datatype, emptyvalue, disabled = \
        select_edit_renderer_props(widget, data)
    optiontags = []
    vocab = attr_value('vocabulary', widget, data, [])
    for key, term in vocabulary(vocab):
        vval = key
        if datatype:
            vval = convert_value_to_datatype(
                key,
                datatype,
                empty_value=emptyvalue
            )
        key = '' if key in [None, UNSET] else key
        attrs = {
            'selected': 'selected' if vval in value else None,
            'value': key,
            'id': cssid(widget, 'input', key),
        }
        if disabled and disabled is not True and vval in disabled:
            attrs['disabled'] = 'disabled'
        optiontags.append(data.tag('option', term, **attrs))
    autofocus = \
        attr_value('autofocus', widget, data) and 'autofocus' or None
    required = attr_value('required', widget, data) and 'required' or None
    block_class = attr_value('block_class', widget, data)
    select_attrs = {
        'name_': widget.dottedpath,
        'id': cssid(widget, 'input'),
        'title': attr_value('title', widget, data) or None,
        'class_': cssclasses(widget, data, additional=[block_class]),
        'multiple': multivalued and 'multiple' or None,
        'size': attr_value('size', widget, data) or None,
        'placeholder': attr_value('placeholder', widget, data) or None,
        'autofocus': autofocus,
        'required': required,
    }
    select_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    select_attrs.update(custom_attrs)
    if disabled is True:
        select_attrs['disabled'] = 'disabled'
    if not optiontags:
        optiontags = [' ']
    rendered = data.tag('select', *optiontags, **select_attrs)
    if multivalued:
        attrs = {
            'type': 'hidden',
            'value': 'exists',
            'name_': '{0}-exists'.format(widget.dottedpath),
            'id': cssid(widget, 'exists'),
        }
        rendered = select_exists_marker(widget, data) + rendered
    return rendered


def select_cb_edit_renderer(widget, data, custom_attrs={}):
    value, multivalued, datatype, emptyvalue, disabled = \
        select_edit_renderer_props(widget, data)
    tags = []
    label_pos = attr_value('listing_label_position', widget, data)
    if label_pos == 'inner':
        # deprecated, use explicit inner-after or inner-before
        label_pos = 'inner-after'
    listing_tag = attr_value('listing_tag', widget, data)
    item_tag = listing_tag == 'div' and 'div' or 'li'
    if multivalued:
        tagtype = 'checkbox'
        wrapper_class = attr_value('checkbox_wrapper_class', widget, data)
        label_class = attr_value('checkbox_label_class', widget, data)
        input_class_additional = attr_value('checkbox_input_class', widget, data)
        # B/C deprecated as of yafowil 2.2
        if not label_class:
            label_class = attr_value('label_checkbox_class', widget, data)
    else:
        tagtype = 'radio'
        wrapper_class = attr_value('radio_wrapper_class', widget, data)
        label_class = attr_value('radio_label_class', widget, data)
        input_class_additional = attr_value('radio_input_class', widget, data)
        # B/C deprecated as of yafowil 2.2
        if not label_class:
            label_class = attr_value('label_radio_class', widget, data)
    vocab = attr_value('vocabulary', widget, data, [])
    for key, term in vocabulary(vocab):
        vval = key
        if datatype:
            vval = convert_value_to_datatype(
                key,
                datatype,
                empty_value=emptyvalue
            )
        key = '' if key in [None, UNSET] else key
        input_attrs = {
            'type': tagtype,
            'value': key,
            'checked': 'checked' if vval in value else None,
            'name_': widget.dottedpath,
            'id': cssid(widget, 'input', key),
            'class_': cssclasses(
                widget,
                data,
                additional=[input_class_additional]
            ),
        }
        if (disabled and disabled is not True and vval in disabled) \
           or disabled is True:
            input_attrs['disabled'] = 'disabled'
        inputtag = data.tag('input', **input_attrs)
        label_attrs = dict(for_=input_attrs['id'], _class=label_class)
        item = generic_positional_rendering_helper(
            'label', term, label_attrs, inputtag, label_pos, data.tag)
        item_wrapper = data.tag(item_tag, item, **{
            'id': cssid(widget, tagtype, key),
            'class': wrapper_class,
        })
        tags.append(item_wrapper)
    wrapper_attrs = {'id': cssid(widget, tagtype, 'wrapper')}
    wrapper_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    wrapper_attrs.update(custom_attrs)
    taglisting = data.tag(listing_tag, *tags, **wrapper_attrs)
    return select_exists_marker(widget, data) + taglisting


@managedprops(
    'data',
    'title',
    'format',
    'vocabulary',
    'multivalued',
    'disabled',
    'listing_label_position',
    'listing_tag',
    'size',
    'block_class',
    'autofocus',
    'placeholder',
    'datatype',
    'emptyvalue',
    'checkbox_wrapper_class',
    'checkbox_label_class',
    'checkbox_input_class',
    'radio_wrapper_class',
    'radio_label_class',
    'radio_input_class',
    *css_managed_props)
def select_edit_renderer(widget, data, custom_attrs={}):
    if attr_value('format', widget, data) == 'block':
        return select_block_edit_renderer(
            widget,
            data,
            custom_attrs=custom_attrs
        )
    return select_cb_edit_renderer(widget, data, custom_attrs=custom_attrs)


@managedprops('data', 'template', 'class', 'multivalued')
def select_display_renderer(widget, data):
    value = fetch_value(widget, data)
    if type(value) in ITER_TYPES and not value:
        value = u''
    multivalued = attr_value('multivalued', widget, data)
    vocab = dict(attr_value('vocabulary', widget, data, []))
    if not multivalued or not value:
        value = vocab.get(value, value)
        if data.tag.translate:
            value = data.tag.translate(value)
        return generic_display_renderer(widget, data, value=value)
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(attr_value('class', widget, data))
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    content = u''
    if multivalued and isinstance(value, STR_TYPE):
        value = [value]
    for key in value:
        content += data.tag('li', vocab.get(key, key))
    return data.tag('ul', content, **attrs)


factory.register(
    'select',
    extractors=[
        select_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
    ],
    edit_renderers=[select_edit_renderer],
    display_renderers=[
        select_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['select'] = """\
Selection Blueprint. Single selection as dropdown or radio-buttons. Multiple
selection as selection-list or as checkboxes.
"""

factory.defaults['select.multivalued'] = None
factory.doc['props']['select.multivalued'] = """\
Flag whether multiple items can be selected.
"""

factory.defaults['select.size'] = None
factory.doc['props']['select.size'] = """\
Size of input if multivalued and format 'block'.
"""

# maybe callable returning '' for single select and [] for multi select
factory.defaults['select.default'] = UNSET

factory.defaults['select.format'] = 'block'
factory.doc['props']['select.format'] = """\
Every value except 'block' results in either a list of radio buttons or
checkboxes depending on the 'multivalued' property.
"""

factory.defaults['select.class'] = 'select'

factory.defaults['select.block_class'] = None
factory.doc['props']['select.block_class'] = """\
CSS class to render on selection if block format.
"""

factory.defaults['select.checkbox_wrapper_class'] = None
factory.doc['props']['select.checkbox_wrapper_class'] = """\
CSS class to render on checkbox wrapper.
"""

factory.defaults['select.checkbox_label_class'] = None
factory.doc['props']['select.checkbox_label_class'] = """\
CSS class to render on checkbox labels.
"""

factory.defaults['select.label_checkbox_class'] = None
factory.doc['props']['select.label_checkbox_class'] = """\
CSS class to render on checkbox labels.

This property is deprecated and will be remove as of yafowil 2.2. Use
``checkbox_label_class`` instead.
"""

factory.defaults['select.checkbox_input_class'] = None
factory.doc['props']['select.checkbox_input_class'] = """\
CSS class to render on checkbox input tag.
"""

factory.defaults['select.radio_wrapper_class'] = None
factory.doc['props']['select.radio_wrapper_class'] = """\
CSS class to render on radio button wrapper.
"""

factory.defaults['select.radio_label_class'] = None
factory.doc['props']['select.radio_label_class'] = """\
CSS class to render on radio button labels.
"""

factory.defaults['select.label_radio_class'] = None
factory.doc['props']['select.label_radio_class'] = """\
CSS class to render on radio button labels.

This property is deprecated and will be remove as of yafowil 2.2. Use
``radio_label_class`` instead.
"""

factory.defaults['select.radio_input_class'] = None
factory.doc['props']['select.radio_input_class'] = """\
CSS class to render on radio button input tag.
"""

factory.defaults['select.listing_tag'] = 'div'
factory.doc['props']['select.listing_tag'] = """\
Desired rendering tag for selection if selection format is 'single'. Valid
values are 'div' and 'ul'.
"""

factory.defaults['select.listing_label_position'] = 'inner-after'
factory.doc['props']['select.listing_label_position'] = """\
Label position if format is 'single'. Behaves the same way as label widget
position property.
"""

factory.doc['props']['select.vocabulary'] = """\
Vocabulary to be used for the selection list. Expects a dict-like or an
iterable or a callable which returns one of both first. An iterable can consist
out of strings or out of tuples with ``(key, value)``. The items in the result
list are in the same order like the vocabulary.
"""

factory.doc['props']['select.disabled'] = """\
Disables the whole widget or single selections. To disable the whole widget
set the value to 'True'. To disable single selection pass a iterable of keys to
disable, i.e. ``['foo', 'baz']``. Defaults to False.
"""

factory.defaults['select.persist'] = True



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import factory
from yafowil.base import fetch_value
from yafowil.common import generic_required_extractor
from yafowil.common import input_attributes_common
from yafowil.tsf import _
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssid
from yafowil.utils import managedprops
from yafowil.utils import vocabulary


###############################################################################
# checkbox
###############################################################################

@managedprops('format')
def checkbox_extractor(widget, data):
    """Extracts data from a single input with type checkbox.
    """
    if '{0}-exists'.format(widget.dottedpath) not in data.request:
        return UNSET
    fmt = attr_value('format', widget, data)
    if fmt == 'bool':
        return widget.dottedpath in data.request
    elif fmt == 'string':
        return data.request.get(widget.dottedpath, '')
    raise ValueError(
        "Checkbox widget has invalid format '{0}' set".format(fmt)
    )


@managedprops(
    'data',
    'title',
    'size',
    'disabled',
    'autofocus',
    'format',
    'disabled',
    'checked',
    'with_label',
    *css_managed_props)
def checkbox_edit_renderer(widget, data):
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = 'checkbox'
    checked = attr_value('checked', widget, data)
    if checked is not None:
        if checked:
            input_attrs['checked'] = 'checked'
    else:
        input_attrs['checked'] = input_attrs['value'] and 'checked' or None
    if attr_value('format', widget, data) == 'bool':
        input_attrs['value'] = ''
    checkbox = tag('input', **input_attrs)
    if attr_value('with_label', widget, data):
        checkbox += tag(
            'label',
            '&nbsp;',
            for_=cssid(widget, 'input'),
            class_='checkbox_label'
        )
    input_attrs = {
        'type': 'hidden',
        'value': 'checkboxexists',
        'name_': "{0}-exists".format(widget.dottedpath),
        'id': cssid(widget, 'checkboxexists'),
    }
    exists_marker = tag('input', **input_attrs)
    return checkbox + exists_marker


@managedprops('class', 'format', 'vocabulary', 'display_proxy')
def checkbox_display_renderer(widget, data):
    """Generic display renderer to render a value.
    """
    value = fetch_value(widget, data)
    fmt = attr_value('format', widget, data)
    if fmt == 'string' and bool(value):
        content = value
    else:
        vocab = dict(vocabulary(attr_value('vocabulary', widget, data, [])))
        # XXX: value might be 'True' Looks odd.
        content = vocab[bool(value)]
        if data.tag.translate:
            content = data.tag.translate(content)
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(attr_value('class', widget, data))
    }
    if attr_value('display_proxy', widget, data):
        widget.attrs['type'] = 'hidden'
        if fmt == 'string':
            input_attrs = input_attributes_common(widget, data, value=value)
            content += data.tag('input', **input_attrs)
        elif bool(value):
            input_attrs = input_attributes_common(widget, data, value='')
            content += data.tag('input', **input_attrs)
        del widget.attrs['type']
        input_attrs = {
            'type': 'hidden',
            'value': 'checkboxexists',
            'name_': "{0}-exists".format(widget.dottedpath),
            'id': cssid(widget, 'checkboxexists'),
        }
        content += data.tag('input', **input_attrs)
    return data.tag('div', content, **attrs)


factory.register(
    'checkbox',
    extractors=[
        checkbox_extractor,
        generic_required_extractor
    ],
    edit_renderers=[checkbox_edit_renderer],
    display_renderers=[checkbox_display_renderer]
)

factory.doc['blueprint']['checkbox'] = """\
Single checkbox blueprint.
"""

factory.defaults['checkbox.default'] = False

factory.defaults['checkbox.format'] = 'bool'
factory.doc['props']['checkbox.format'] = """\
Data-type of the extracted value. One out of ``bool`` or ``string``.
"""

factory.defaults['checkbox.class'] = 'checkbox'

factory.defaults['checkbox.disabled'] = False
factory.doc['props']['checkbox.disabled'] = """\
Flag whether checkbox is disabled.
"""

factory.defaults['checkbox.checked'] = None
factory.doc['props']['checkbox.checked'] = """\
Set 'checked' attribute explicit. If not given, compute by value.
"""

factory.defaults['checkbox.vocabulary'] = {
    True: _('yes', default=u'Yes'),
    False: _('no', default=u'No'),
    UNSET: _('unset', default=u'Unset'),  # XXX: never used right now?
}

factory.doc['props']['checkbox.vocabulary'] = """\
In display mode and if ```bool``` is set to ```True``` this mapping will be
used for display of the value. Expected keys are ```True```, ```False``` and
```UNSET```.
"""

factory.defaults['checkbox.with_label'] = False
factory.doc['props']['checkbox.with_label'] = """\
Render empty label tag after visible checkbox in order to make checkbox UI
customizable via CSS like so::

    input.large_checkbox {
        display: none;
    }
    input.large_checkbox + label {
        width: 59px;
        height: 60px;
        background: url('/checkbox_large.png');
    }
    input.large_checkbox:checked + label {
        background: url('/checkbox_large_selected.png');
    }
"""

factory.defaults['checkbox.required_class'] = 'required'

factory.defaults['checkbox.persist'] = True



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.common import input_generic_renderer
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.tsf import _
import re


###############################################################################
# url
###############################################################################

URL_RE = r'^(ftp|http|https):\/\/(\w+:{0,1}\w*@)?(\S+)(:[0-9]+)?(\/|\/([\w#!:'
URL_RE += r'.?+=&%@!\-\/]))?$'


def url_extractor(widget, data):
    val = data.extracted
    if not val:
        return val
    if not re.match(URL_RE, val is not UNSET and val or ''):
        message = _('web_address_not_valid',
                    default=u'Input not a valid web address.')
        raise ExtractionError(message)
    return val


factory.register(
    'url',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        url_extractor
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['url'] = """\
URL aka web address (HTML5) input blueprint.
"""

factory.defaults['url.type'] = 'url'

factory.defaults['url.default'] = ''

factory.defaults['url.required_class'] = 'required'

factory.defaults['url.class'] = 'url'

factory.defaults['url.persist'] = True



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.base import fetch_value
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.compat import ITER_TYPES
from yafowil.datatypes import generic_datatype_extractor
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.textarea import textarea_attributes
from yafowil.textarea import textarea_managed_props
from yafowil.tsf import _
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import cssid
from yafowil.utils import managedprops


###############################################################################
# lines
###############################################################################

def lines_extractor(widget, data):
    """Extract textarea value as list of lines.
    """
    extracted = data.extracted
    if not extracted:
        return list()
    return extracted.split('\n')


@managedprops(*textarea_managed_props)
def lines_edit_renderer(widget, data):
    """Renders text area with list value as lines.
    """
    tag = data.tag
    area_attrs = textarea_attributes(widget, data)
    value = fetch_value(widget, data)
    if value is None:
        value = u''
    else:
        value = u'\n'.join(value)
    return tag('textarea', value, **area_attrs)


@managedprops('class', 'data')
def lines_display_renderer(widget, data):
    value = fetch_value(widget, data)
    if type(value) in ITER_TYPES and not value:
        value = u''
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(attr_value('class', widget, data))
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    content = u''
    for line in value:
        content += data.tag('li', line)
    return data.tag('ul', content, **attrs)


factory.register(
    'lines',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        lines_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
    ],
    edit_renderers=[lines_edit_renderer],
    display_renderers=[
        lines_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['lines'] = """\
Lines blueprint. Renders a textarea and extracts lines as list.
"""

factory.defaults['lines.default'] = ''

factory.defaults['lines.class'] = 'lines'

factory.defaults['lines.wrap'] = None
factory.doc['props']['lines.wrap'] = """\
Either ``soft``, ``hard``, ``virtual``, ``physical`` or  ``off``.
"""

factory.defaults['lines.cols'] = 40
factory.doc['props']['lines.cols'] = """\
Number of characters.
"""

factory.defaults['lines.rows'] = 8
factory.doc['props']['lines.rows'] = """\
Number of lines.
"""

factory.defaults['lines.readonly'] = None
factory.doc['props']['lines.readonly'] = """\
Flag textarea is readonly.
"""

factory.defaults['lines.persist'] = True



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.common import input_generic_renderer
from yafowil.datatypes import generic_datatype_extractor
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.tsf import _
from yafowil.utils import attr_value
from yafowil.utils import managedprops


###############################################################################
# number
###############################################################################

@managedprops('min', 'max', 'step')
def number_extractor(widget, data):
    val = data.extracted
    if val is UNSET:
        return val
    min_val = attr_value('min', widget, data)
    if min_val is not None and val < min_val:
        message = _('input_number_minimum_value',
                    default=u'Value has to be at minimum ${min}.',
                    mapping={'min': min_val})
        raise ExtractionError(message)
    max_val = attr_value('max', widget, data)
    if max_val is not None and val > max_val:
        message = _('input_number_maximum_value',
                    default=u'Value has to be at maximum ${max}.',
                    mapping={'max': max_val})
        raise ExtractionError(message)
    step = attr_value('step', widget, data)
    if step:
        minimum = min_val or 0
        if (val - minimum) % step:
            if minimum:
                message = _(
                    'input_number_step_and_minimum_value',
                    default=u'Value ${val} has to be in stepping of ${step} '
                            u'based on a floor value of ${minimum}',
                    mapping={
                        'val': val,
                        'step': step,
                        'minimum': minimum,
                    })
            else:
                message = _(
                    'input_number_step_value',
                    default=u'Value ${val} has to be in stepping of ${step}',
                    mapping={
                        'val': val,
                        'step': step,
                    })
            raise ExtractionError(message)
    return val


factory.register(
    'number',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
        number_extractor,
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['number'] = """\
Number blueprint (HTML5).
"""

factory.defaults['number.type'] = 'number'

factory.defaults['number.default'] = ''

factory.defaults['number.emptyvalue'] = UNSET

factory.defaults['number.datatype'] = float
factory.doc['props']['number.datatype'] = """\
Callable for converting extracted value to output datatype. Allowed datatypes
are ``int`` and ``float``

``datatype`` can also be defined as string with value out of `'int'``,
``'integer'`` or ``'float'``.
"""

factory.defaults['number.allowed_datatypes'] = [
    int, float,
    # B/C
    'int', 'integer', 'float'
]

factory.defaults['number.min'] = None
factory.doc['props']['number.min'] = """\
Minimum value.
"""

factory.defaults['number.max'] = None
factory.doc['props']['number.max'] = """\
Maximum value.
"""

factory.defaults['number.step'] = None
factory.doc['props']['number.step'] = """\
Stepping value must be in.
"""

factory.defaults['number.required_class'] = 'required'

factory.defaults['number.class'] = 'number'

factory.defaults['number.persist'] = True



from __future__ import print_function
from importlib import reload
from node.tests import NodeTestCase
from yafowil import button
from yafowil import checkbox
from yafowil import common
from yafowil import compound
from yafowil import datatypes
from yafowil import email
from yafowil import field
from yafowil import file
from yafowil import hidden
from yafowil import lines
from yafowil import number
from yafowil import password
from yafowil import persistence
from yafowil import proxy
from yafowil import search
from yafowil import select
from yafowil import table
from yafowil import tag as tag_module
from yafowil import text
from yafowil import textarea
from yafowil import url
from yafowil.base import factory
from yafowil.utils import Tag
import lxml.etree as etree
import sys
import unittest


class YafowilTestCase(NodeTestCase):

    def setUp(self):
        super(YafowilTestCase, self).setUp()
        factory.push_state()
        factory.clear()
        reload(button)
        reload(checkbox)
        reload(common)
        reload(compound)
        reload(datatypes)
        reload(email)
        reload(field)
        reload(file)
        reload(hidden)
        reload(lines)
        reload(number)
        reload(password)
        reload(persistence)
        reload(proxy)
        reload(search)
        reload(select)
        reload(table)
        reload(tag_module)
        reload(text)
        reload(textarea)
        reload(url)

    def tearDown(self):
        factory.pop_state()
        super(YafowilTestCase, self).tearDown()


def fxml(xml):
    et = etree.fromstring(xml)
    return etree.tostring(et, pretty_print=True).decode('utf-8')


def wrapped_fxml(value):
    return fxml(u'<div>' + value + u'</div>')


tag = Tag(lambda msg: msg)


def test_suite():
    from yafowil.tests import test_base
    from yafowil.tests import test_button
    from yafowil.tests import test_checkbox
    from yafowil.tests import test_common
    from yafowil.tests import test_compound
    from yafowil.tests import test_controller
    from yafowil.tests import test_datatypes
    from yafowil.tests import test_email
    from yafowil.tests import test_field
    from yafowil.tests import test_file
    from yafowil.tests import test_hidden
    from yafowil.tests import test_lines
    from yafowil.tests import test_number
    from yafowil.tests import test_password
    from yafowil.tests import test_persistence
    from yafowil.tests import test_proxy
    from yafowil.tests import test_resources
    from yafowil.tests import test_search
    from yafowil.tests import test_select
    from yafowil.tests import test_table
    from yafowil.tests import test_tag
    from yafowil.tests import test_text
    from yafowil.tests import test_textarea
    from yafowil.tests import test_tsf
    from yafowil.tests import test_url
    from yafowil.tests import test_utils

    suite = unittest.TestSuite()

    suite.addTest(unittest.findTestCases(test_base))
    suite.addTest(unittest.findTestCases(test_button))
    suite.addTest(unittest.findTestCases(test_checkbox))
    suite.addTest(unittest.findTestCases(test_common))
    suite.addTest(unittest.findTestCases(test_compound))
    suite.addTest(unittest.findTestCases(test_controller))
    suite.addTest(unittest.findTestCases(test_datatypes))
    suite.addTest(unittest.findTestCases(test_email))
    suite.addTest(unittest.findTestCases(test_field))
    suite.addTest(unittest.findTestCases(test_file))
    suite.addTest(unittest.findTestCases(test_hidden))
    suite.addTest(unittest.findTestCases(test_lines))
    suite.addTest(unittest.findTestCases(test_number))
    suite.addTest(unittest.findTestCases(test_password))
    suite.addTest(unittest.findTestCases(test_persistence))
    suite.addTest(unittest.findTestCases(test_proxy))
    suite.addTest(unittest.findTestCases(test_resources))
    suite.addTest(unittest.findTestCases(test_search))
    suite.addTest(unittest.findTestCases(test_select))
    suite.addTest(unittest.findTestCases(test_table))
    suite.addTest(unittest.findTestCases(test_tag))
    suite.addTest(unittest.findTestCases(test_text))
    suite.addTest(unittest.findTestCases(test_textarea))
    suite.addTest(unittest.findTestCases(test_tsf))
    suite.addTest(unittest.findTestCases(test_url))
    suite.addTest(unittest.findTestCases(test_utils))

    return suite


def run_tests():
    from zope.testrunner.runner import Runner

    runner = Runner(found_suites=[test_suite()])
    runner.run()
    sys.exit(int(runner.failed))


if __name__ == '__main__':
    run_tests()



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.compat import BYTES_TYPE
from yafowil.compat import IS_PY2
from yafowil.compat import LONG_TYPE
from yafowil.compat import STR_TYPE
from yafowil.compat import UNICODE_TYPE
from yafowil.tsf import _
from yafowil.utils import EMPTY_VALUE
from yafowil.utils import attr_value
from yafowil.utils import managedprops
import codecs
import uuid
import warnings


@managedprops('emptyvalue')
def generic_emptyvalue_extractor(widget, data):
    """Return emptyvalue if widget present in request and raw value is empty.
    """
    try:
        if not data.request[widget.dottedpath]:
            return attr_value('emptyvalue', widget, data, data.extracted)
    except KeyError:
        pass
    return data.extracted


factory.doc['props']['emptyvalue'] = """\
If configured and received value in request is empty, return as extracted
value.
"""


class DatatypeConverter(object):
    """Value conversion for extraction and rendering."""

    def __init__(self, type_=None):
        self.type_ = type_

    def to_value(self, value):
        """Convert given value on extraction to desired type."""
        try:
            if isinstance(value, self.type_):
                return value
        except TypeError:
            # type is class or function
            pass
        return self.type_(value)

    def to_form(self, value):
        """Convert given value to unicode string for rendering."""
        if not isinstance(value, UNICODE_TYPE):
            value = u'{}'.format(value)
        return value.replace('"', '&quot;')


class BytesDatatypeConverter(DatatypeConverter):
    """Datatype converter for bytes."""

    def to_value(self, value):
        if isinstance(value, BYTES_TYPE):
            return value
        # passed unicode value must contain ascii characters only
        return codecs.escape_decode(value.encode('ascii'))[0]

    def to_form(self, value):
        if isinstance(value, UNICODE_TYPE):
            return value
        return codecs.escape_encode(value)[0].decode('ascii')


class UnicodeDatatypeConverter(DatatypeConverter):
    """Datatype converter for unicode."""

    def __init__(self):
        super(UnicodeDatatypeConverter, self).__init__(UNICODE_TYPE)

    def to_value(self, value):
        if isinstance(value, self.type_):
            return value
        return self.type_(value, encoding='utf-8')


class FloatDatatypeConverter(DatatypeConverter):
    """Datatype converter for float."""

    def __init__(self):
        super(FloatDatatypeConverter, self).__init__(float)

    def to_value(self, value):
        if isinstance(value, STR_TYPE):
            value = value.replace(',', '.')
        return self.type_(value)


DATATYPE_CONVERTERS = {
    BYTES_TYPE: BytesDatatypeConverter(),
    UNICODE_TYPE: UnicodeDatatypeConverter(),
    float: FloatDatatypeConverter()
}

# B/C, will be removed as of yafowil 3.2
LEGACY_DATATYPE_CONVERTERS = {
    'str': BytesDatatypeConverter(),
    'unicode': UnicodeDatatypeConverter(),
    'int': int,
    'integer': int,
    'long': LONG_TYPE,  # long only exists in python 2
    'float': FloatDatatypeConverter(),
    'uuid': uuid.UUID
}


def lookup_datatype_converter(datatype):
    """Lookup datatype converter for given datatype.

    Datatype is either a type or an instance of ``DatatypeConverter``.
    If datatype is a type, an instance of ``DatatypeConverter`` gets created.

    For B/C behavior, datatype can be a string out of 'str', 'unicode',
    'int', 'integer', 'long', 'float' or 'uuid'. This behavior is deprecated
    and will be removed as of yafowil 3.2.
    """
    if isinstance(datatype, STR_TYPE):
        warnings.warn(
            'Passing ``datatype`` as string to ``convert_value_to_datatype`` '
            'is deprecated and will be removed as of yafowil 3.2.'
        )
        converter = LEGACY_DATATYPE_CONVERTERS[datatype]
    else:
        converter = DATATYPE_CONVERTERS.get(datatype, datatype)
    if not isinstance(converter, DatatypeConverter):
        converter = DatatypeConverter(converter)
    return converter


def convert_value_to_datatype(value, datatype, empty_value=EMPTY_VALUE):
    """Convert given value to datatype.

    Uses ``lookup_datatype_converter`` to lookup the appropriate datatype
    converter.

    If value is ``UNSET``, return ``UNSET``, regardless of given datatype.

    If value is ``EMPTY_VALUE``, return ``empty_value``, which defaults to
    ``EMPTY_VALUE`` marker.

    If value is ``None`` or ``''``, return ``empty_value``, which defaults to
    ``EMPTY_VALUE`` marker. Be aware that empty value marker is even returned
    if ``str`` datatype, to provide a consistent behavior.

    Converter callables must raise one out of the following exceptions if
    conversion fails:

        * ``ValueError``
        * ``UnicodeDecodeError``
        * ``UnicodeEncodeError``
    """
    if value is UNSET:
        return UNSET
    if value is EMPTY_VALUE:
        return empty_value
    if value in [None, '']:
        return empty_value
    converter = lookup_datatype_converter(datatype)
    return converter.to_value(value)


def convert_values_to_datatype(value, datatype, empty_value=EMPTY_VALUE):
    """Convert given value(s) to datatype.

    If value is list, each item in list gets converted to datatype.

    ``convert_value_to_datatype`` function is used for each value.

    Returns converted value or list of converted values.
    """
    if isinstance(value, list):
        res = list()
        for item in value:
            res.append(convert_value_to_datatype(
                item,
                datatype,
                empty_value=empty_value
            ))
        return res
    return convert_value_to_datatype(value, datatype, empty_value=empty_value)


DATATYPE_LABELS = {
    BYTES_TYPE: _('datatype_str', default='string'),
    UNICODE_TYPE: _('datatype_unicode', default='unicode'),
    int: _('datatype_integer', default='integer'),
    float: _('datatype_float', default='floating point number'),
    uuid.UUID: _('datatype_uuid', default='UUID')
}
if IS_PY2:  # pragma: no cover
    DATATYPE_LABELS[LONG_TYPE] = _('datatype_long', default='long integer')
else:  # pragma: no cover
    DATATYPE_LABELS[LONG_TYPE] = DATATYPE_LABELS[int]
# B/C, will be removed as of yafowil 3.2
DATATYPE_LABELS['str'] = DATATYPE_LABELS[BYTES_TYPE]
DATATYPE_LABELS['unicode'] = DATATYPE_LABELS[UNICODE_TYPE]
DATATYPE_LABELS['int'] = DATATYPE_LABELS[int]
DATATYPE_LABELS['integer'] = DATATYPE_LABELS[int]
DATATYPE_LABELS['long'] = DATATYPE_LABELS[LONG_TYPE]
DATATYPE_LABELS['float'] = DATATYPE_LABELS[float]
DATATYPE_LABELS['uuid'] = DATATYPE_LABELS[uuid.UUID]


@managedprops(
    'datatype',
    'allowed_datatypes',
    'datatype_message',
    'emptyvalue')
def generic_datatype_extractor(widget, data):
    """Convert extracted value to ``datatype``.

    If extracted value is ``UNSET`` return ``UNSET``.

    If no ``datatype`` given, return extracted value.

    Otherwise try to convert value to given ``datatype`` and return the
    converted value or raise an ``ExtractionError`` if conversion fails.

    Extraction error message can be customized with ``datatype_message``
    property. Value can also be a list, then all items inside the list are
    converted.
    """
    extracted = data.extracted
    if extracted is UNSET:
        return extracted
    # datatype is one of the rare cases where an attribute callable not follows
    # the typical signature taking widget and data as arguments, but is just
    # called with the value.
    datatype = widget.attrs.get('datatype', None)
    if not datatype:
        return extracted
    allowed_datatypes = attr_value('allowed_datatypes', widget, data)
    if allowed_datatypes and datatype not in allowed_datatypes:
        raise ValueError('Datatype not allowed: "{0}"'.format(datatype))
    try:
        emptyvalue = attr_value('emptyvalue', widget, data, EMPTY_VALUE)
        return convert_values_to_datatype(
            extracted,
            datatype,
            empty_value=emptyvalue
        )
    except KeyError:
        raise ValueError('Datatype unknown: "{0}"'.format(datatype))
    except (ValueError, UnicodeEncodeError, UnicodeDecodeError):
        datatype_message = attr_value('datatype_message', widget, data)
        if not datatype_message:
            datatype_label = DATATYPE_LABELS.get(datatype)
            if not datatype_label:
                datatype_message = _(
                    'generic_datatype_message',
                    default=u'Input conversion failed.'
                )
            else:
                if data.tag.translate:
                    datatype_label = data.tag.translate(datatype_label)
                datatype_message = _(
                    'standard_datatype_message',
                    default=u'Input is not a valid ${datatype}.',
                    mapping={
                        'datatype': datatype_label
                    }
                )
        raise ExtractionError(datatype_message)


factory.defaults['datatype'] = None
factory.doc['props']['datatype'] = """\
Datatype is either a type or an instance of ``DatatypeConverter``.
If datatype is a type, an instance of ``DatatypeConverter`` gets created.

For B/C behavior, datatype can be a string out of 'str', 'unicode',
'int', 'integer', 'long', 'float' or 'uuid'. This behavior is deprecated
and will be removed as of yafowil 3.2.

**NOTE**: As of Python 3, string identifiers should not be used at all, but if
you do so, the following rules apply:
    * 'str' maps to ``bytes`` type.
    * 'unicode' maps to ``str`` type.
    * 'long' maps to ``int`` type.

Custom datatype converters must raise one out of the following exceptions if
conversion fails:
    * ``ValueError``
    * ``UnicodeDecodeError``
    * ``UnicodeEncodeError``
"""

factory.defaults['allowed_datatypes'] = UNSET
factory.doc['props']['allowed_datatypes'] = """\
List of allowed datatypes. If ``UNSET``, datatype converters are not
restricted.
"""

factory.defaults['datatype_message'] = None
factory.doc['props']['datatype_message'] = """\
Custom extraction error message if ``datatype`` conversion fails.
"""



from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase
from yafowil.utils import EMPTY_VALUE


class TestProxy(YafowilTestCase):

    def test_proxy_blueprint(self):
        # Used to pass hidden arguments out of form namespace
        widget = factory(
            'proxy',
            name='PROXY',
            value='1')
        self.assertEqual(widget(), (
            '<input id="input-PROXY" name="PROXY" type="hidden" value="1" />'
        ))
        self.assertEqual(widget(request={'PROXY': '2'}), (
            '<input id="input-PROXY" name="PROXY" type="hidden" value="2" />'
        ))

        # Emptyvalue
        widget = factory(
            'proxy',
            name='PROXY',
            value='',
            props={
                'emptyvalue': '1.0'
            })
        data = widget.extract(request={'PROXY': ''})
        self.assertEqual(data.name, 'PROXY')
        self.assertEqual(data.value, '')
        self.assertEqual(data.extracted, '1.0')
        self.assertEqual(data.errors, [])

        # Datatype
        widget = factory(
            'proxy',
            name='PROXY',
            value='',
            props={
                'emptyvalue': '1.0',
                'datatype': float
            })
        data = widget.extract(request={'PROXY': '2.0'})
        self.assertEqual(data.name, 'PROXY')
        self.assertEqual(data.value, '')
        self.assertEqual(data.extracted, 2.0)
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'PROXY': ''})
        self.assertEqual(data.name, 'PROXY')
        self.assertEqual(data.value, '')
        self.assertEqual(data.extracted, 1.0)
        self.assertEqual(data.errors, [])

        # Default emptyvalue extraction
        del widget.attrs['emptyvalue']
        data = widget.extract(request={'PROXY': ''})
        self.assertEqual(data.name, 'PROXY')
        self.assertEqual(data.value, '')
        self.assertEqual(data.extracted, EMPTY_VALUE)
        self.assertEqual(data.errors, [])

        # Persist defaults to false
        widget = factory(
            'proxy',
            name='PROXY',
            props={
                'persist_writer': write_mapping_writer
            })
        data = widget.extract(request={'PROXY': '10'})
        model = dict()
        data.write(model)
        self.assertEqual(model, {})

        # If proxy widgets really need to be persisted, ``persist`` property
        # needs to be set explicitely
        widget.attrs['persist'] = True
        data = widget.extract(request={'PROXY': '10'})
        data.write(model)
        self.assertEqual(model, {'PROXY': '10'})



from node.utils import UNSET
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase


class TestTextarea(YafowilTestCase):

    def test_textarea_blueprint(self):
        # Textarea widget
        widget = factory(
            'textarea',
            name='MYTEXTAREA',
            value=None)
        self.assertEqual(widget(), (
            '<textarea class="textarea" cols="80" id="input-MYTEXTAREA" '
            'name="MYTEXTAREA" rows="25"></textarea>'
        ))

        widget = factory(
            'textarea',
            name='MYTEXTAREA',
            value=None,
            props={
                'data': {
                    'foo': 'bar'
                },
            })
        self.assertEqual(widget(), (
            '<textarea class="textarea" cols="80" data-foo=\'bar\' '
            'id="input-MYTEXTAREA" name="MYTEXTAREA" rows="25"></textarea>'
        ))

        widget.mode = 'display'
        self.assertEqual(widget(), (
            '<div class="display-textarea" data-foo=\'bar\' '
            'id="display-MYTEXTAREA"></div>'
        ))

        # Emptyvalue
        widget = factory(
            'textarea',
            name='MYTEXTAREA',
            props={
                'emptyvalue': 'EMPTYVALUE',
            })
        data = widget.extract(request={'MYTEXTAREA': ''})
        self.assertEqual(data.name, 'MYTEXTAREA')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'EMPTYVALUE')
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'MYTEXTAREA': 'NOEMPTY'})
        self.assertEqual(data.name, 'MYTEXTAREA')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'NOEMPTY')
        self.assertEqual(data.errors, [])

        # Persist
        widget = factory(
            'textarea',
            name='MYTEXTAREA',
            props={
                'persist_writer': write_mapping_writer
            })
        data = widget.extract(request={'MYTEXTAREA': 'Text'})
        model = dict()
        data.write(model)
        self.assertEqual(model, {'MYTEXTAREA': 'Text'})



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.common import input_attributes_common
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.tsf import _
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import managedprops
import re


###############################################################################
# password
###############################################################################

@managedprops('minlength')
def minlength_extractor(widget, data):
    """Validate minlength of a string input.

    Only perform if ``minlength`` property is set.

    Properties:

    ``minlength``
        Minimum length of string as int.
    """
    val = data.extracted
    if val is UNSET:
        return val
    minlength = attr_value('minlength', widget, data, -1)
    if minlength != -1:
        if len(val) < minlength:
            raise ExtractionError(_(
                'minlength_extraction_error',
                default=u'Input must have at least ${len} characters.',
                mapping={'len': minlength}
            ))
    return val


@managedprops('ascii')
def ascii_extractor(widget, data):
    """Validate if a string is ASCII encoding.

    Only perform if ``ascii`` property evaludates to True.

    Properties:

    ``ascii``
        Flag  ascii check should perform.
    """
    val = data.extracted
    if val is UNSET:
        return val
    if not attr_value('ascii', widget, data, False):
        return val
    try:
        val.encode('ascii')
    except UnicodeEncodeError:
        raise ExtractionError(_(
            u'ascii_extractor_error',
            default=u'Input contains illegal characters.'
        ))
    return val


LOWER_CASE_RE = r'(?=.*[a-z])'
UPPER_CASE_RE = r'(?=.*[A-Z])'
DIGIT_RE = r'(?=.*[\d])'
SPECIAL_CHAR_RE = r'(?=.*[\W])'
RE_PASSWORD_ALL = [
    LOWER_CASE_RE,
    UPPER_CASE_RE,
    DIGIT_RE,
    SPECIAL_CHAR_RE]
PASSWORD_NOCHANGE_VALUE = '_NOCHANGE_'


@managedprops('strength', 'weak_password_message')
def password_extractor(widget, data):
    """Extract and validate password input.

    If extracted password is unchanged, return ``UNSET``. Consider this when
    reading from password widgets!

    This extractor provides a strength check. It only performs if ``strenght``
    property is set. Strength check is done by four rules:
        - input contains lowercase character
        - input contains uppercase character
        - input contains digit
        - input contains special character.

    Properties:

    ``strength``
        Integer value <= 4. Define how many rules must apply to consider a
        password valid.
    """
    val = data.extracted
    if val == PASSWORD_NOCHANGE_VALUE:
        return UNSET
    if val is UNSET:
        return val
    required_strength = attr_value('strength', widget, data, 0)
    if required_strength <= 0:
        return val
    if required_strength > len(RE_PASSWORD_ALL):
        required_strength = len(RE_PASSWORD_ALL)
    strength = 0
    for reg_exp in RE_PASSWORD_ALL:
        if re.match(reg_exp, val):
            strength += 1
    if strength < required_strength:
        error = attr_value('weak_password_message', widget, data)
        raise ExtractionError(error)
    return val


def _pwd_value(widget, data):
    if data.extracted is not UNSET:
        return data.extracted
    if data.value is not UNSET \
       and data.value is not None:
        return PASSWORD_NOCHANGE_VALUE
    return attr_value('default', widget, data)


@managedprops(
    'data',
    'title',
    'size',
    'disabled',
    'autofocus',
    'placeholder',
    'autocomplete',
    *css_managed_props)
def password_edit_renderer(widget, data):
    """Render password widget.
    """
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = 'password'
    input_attrs['value'] = _pwd_value(widget, data)
    return tag('input', **input_attrs)


@managedprops('displayplaceholder')
def password_display_renderer(widget, data):
    value = _pwd_value(widget, data)
    if value == PASSWORD_NOCHANGE_VALUE:
        return attr_value('displayplaceholder', widget, data)
    return u''


factory.register(
    'password',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        minlength_extractor,
        ascii_extractor,
        password_extractor
    ],
    edit_renderers=[password_edit_renderer],
    display_renderers=[password_display_renderer]
)

factory.doc['blueprint']['password'] = """\
Password blueprint.

The password is never rendered to markup, instead
``yafowil.common.PASSWORD_NOCHANGE_VALUE`` is set as ``value`` property on
dom element. See ``yafowil.common.password_extractor`` for details on
password extraction.
"""

factory.defaults['password.required_class'] = 'required'

factory.defaults['password.default'] = ''

factory.defaults['password.class'] = 'password'

factory.defaults['password.minlength'] = -1
factory.doc['props']['password.size'] = """\
Maximum length of password.
"""

factory.doc['props']['password.minlength'] = """\
Minimum length of password.
"""

factory.defaults['password.ascii'] = False
factory.doc['props']['password.ascii'] = """\
Flag ascii check should performed.
"""

factory.defaults['password.strength'] = -1
factory.doc['props']['password.strength'] = """\
Integer value <= 4. Define how many rules must apply to consider a password
valid.
"""

factory.defaults['weak_password_message'] = _('weak_password_message',
                                              default=u'Password too weak')
factory.doc['props']['password.weak_password_message'] = """\
Message shown if password is not strong enough.
"""

factory.defaults['password.displayplaceholder'] = u'*' * 8
factory.doc['props']['password.displayplaceholder'] = """\
Placeholder shown in display mode if password was set.
"""

factory.defaults['password.persist'] = True


from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.compat import UNICODE_TYPE
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase
from yafowil.tests import fxml
from yafowil.tests import wrapped_fxml
from yafowil.utils import EMPTY_VALUE
import uuid


class TestSelect(YafowilTestCase):

    def test_select_blueprint_single_value(self):
        # Default single value selection
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'vocabulary': vocab
            })
        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option id="input-MYSELECT-one" selected="selected"
                  value="one">One</option>
          <option id="input-MYSELECT-two" value="two">Two</option>
          <option id="input-MYSELECT-three" value="three">Three</option>
          <option id="input-MYSELECT-four" value="four">Four</option>
        </select>
        """, fxml(widget()))

        data = widget.extract({'MYSELECT': 'two'})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 'two')

        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option id="input-MYSELECT-one" value="one">One</option>
          <option id="input-MYSELECT-two" selected="selected"
                  value="two">Two</option>
          <option id="input-MYSELECT-three" value="three">Three</option>
          <option id="input-MYSELECT-four" value="four">Four</option>
        </select>
        """, fxml(widget(data=data)))

        # Single value selection completly disabled
        widget.attrs['disabled'] = True
        self.checkOutput("""
        <select class="select" disabled="disabled" id="input-MYSELECT"
                name="MYSELECT">
          <option id="input-MYSELECT-one" selected="selected"
                  value="one">One</option>
          <option id="input-MYSELECT-two" value="two">Two</option>
          <option id="input-MYSELECT-three" value="three">Three</option>
          <option id="input-MYSELECT-four" value="four">Four</option>
        </select>
        """, fxml(widget()))

        # Single value selection with specific options disabled
        widget.attrs['disabled'] = ['two', 'four']
        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option id="input-MYSELECT-one" selected="selected"
                  value="one">One</option>
          <option disabled="disabled" id="input-MYSELECT-two"
                  value="two">Two</option>
          <option id="input-MYSELECT-three" value="three">Three</option>
          <option disabled="disabled" id="input-MYSELECT-four"
                  value="four">Four</option>
        </select>
        """, fxml(widget()))

        del widget.attrs['disabled']

        # Single value selection display mode
        widget.mode = 'display'
        self.assertEqual(
            widget(),
            '<div class="display-select" id="display-MYSELECT">One</div>'
        )

        widget.attrs['display_proxy'] = True
        self.checkOutput("""
        <div>
          <div class="display-select" id="display-MYSELECT">One</div>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="one"/>
        </div>
        """, wrapped_fxml(widget()))

        data = widget.extract(request={'MYSELECT': 'two'})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'one')
        self.assertEqual(data.extracted, 'two')
        self.assertEqual(data.errors, [])

        self.checkOutput("""
        <div>
          <div class="display-select" id="display-MYSELECT">Two</div>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="two"/>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Single value selection with datatype set
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'datatype': uuid.UUID
            })

        # Preselected values
        widget.getter = UNSET
        widget.attrs['vocabulary'] = [
            (UNSET, 'Empty value'),
            (uuid.UUID('1b679ef8-9068-45f5-8bb8-4007264aa7f7'), 'One')
        ]
        res = widget()
        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option id="input-MYSELECT-" selected="selected"
                  value="">Empty value</option>
          <option id="input-MYSELECT-1b679ef8-..."
                  value="1b679ef8-...">One</option>
        </select>
        """, fxml(res))

        widget.getter = EMPTY_VALUE
        widget.attrs['vocabulary'][0] = (EMPTY_VALUE, 'Empty value')
        self.assertEqual(res, widget())

        widget.getter = None
        widget.attrs['vocabulary'][0] = (None, 'Empty value')
        self.assertEqual(res, widget())

        widget.getter = ''
        widget.attrs['vocabulary'][0] = ('', 'Empty value')
        self.assertEqual(res, widget())

        widget.getter = uuid.UUID('1b679ef8-9068-45f5-8bb8-4007264aa7f7')
        res = widget()
        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option id="input-MYSELECT-"
                  value="">Empty value</option>
          <option id="input-MYSELECT-1b679ef8-..."
                  selected="selected" value="1b679ef8-...">One</option>
        </select>
        """, fxml(res))

        # Note, vocabulary keys are converted to ``datatype`` while widget
        # value needs to be of type defined in ``datatype`` or one from the
        # valid empty values
        widget.attrs['vocabulary'] = [
            (UNSET, 'Empty value'),
            ('1b679ef8-9068-45f5-8bb8-4007264aa7f7', 'One')
        ]
        res = widget()
        self.assertEqual(res, widget())

        # Test ``datatype`` extraction with selection
        vocab = [
            (EMPTY_VALUE, 'Empty value'),
            (1, 'One'),
            (2, 'Two'),
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value=2,
            props={
                'vocabulary': vocab,
                'datatype': int
            })
        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option id="input-MYSELECT-" value="">Empty value</option>
          <option id="input-MYSELECT-1" value="1">One</option>
          <option id="input-MYSELECT-2" selected="selected"
                  value="2">Two</option>
        </select>
        """, fxml(widget()))

        data = widget.extract({'MYSELECT': ''})
        self.assertEqual(data.extracted, EMPTY_VALUE)

        data = widget.extract({'MYSELECT': '1'})
        self.assertEqual(data.extracted, 1)

        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option id="input-MYSELECT-" value="">Empty value</option>
          <option id="input-MYSELECT-1" selected="selected"
                  value="1">One</option>
          <option id="input-MYSELECT-2" value="2">Two</option>
        </select>
        """, fxml(widget(data=data)))

        # Test extraction with ``emptyvalue`` set
        widget.attrs['emptyvalue'] = UNSET
        data = widget.extract({})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYSELECT': ''})
        self.assertEqual(data.extracted, UNSET)

        widget.attrs['emptyvalue'] = None
        data = widget.extract({})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYSELECT': ''})
        self.assertEqual(data.extracted, None)

        widget.attrs['emptyvalue'] = 0
        data = widget.extract({})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYSELECT': ''})
        self.assertEqual(data.extracted, 0)

        # Single value selection with ``datatype`` set completly disabled
        widget.attrs['disabled'] = True
        self.checkOutput("""
        <select class="select" disabled="disabled" id="input-MYSELECT"
                name="MYSELECT">
          <option id="input-MYSELECT-" value="">Empty value</option>
          <option id="input-MYSELECT-1" value="1">One</option>
          <option id="input-MYSELECT-2" selected="selected"
                  value="2">Two</option>
        </select>
        """, fxml(widget()))

        # Single value selection with ``datatype`` with specific options
        # disabled
        widget.attrs['emptyvalue'] = None
        widget.attrs['disabled'] = [None, 2]
        rendered = widget()
        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT">
          <option disabled="disabled" id="input-MYSELECT-"
                  value="">Empty value</option>
          <option id="input-MYSELECT-1" value="1">One</option>
          <option disabled="disabled" id="input-MYSELECT-2"
                  selected="selected" value="2">Two</option>
        </select>
        """, fxml(rendered))

        widget.attrs['emptyvalue'] = UNSET
        widget.attrs['disabled'] = [UNSET, 2]
        self.assertEqual(widget(), rendered)

        widget.attrs['emptyvalue'] = EMPTY_VALUE
        widget.attrs['disabled'] = [EMPTY_VALUE, 2]
        self.assertEqual(widget(), rendered)

        widget.attrs['emptyvalue'] = 0
        widget.attrs['disabled'] = [0, 2]
        self.assertEqual(widget(), rendered)

        del widget.attrs['disabled']

        # Single value selection with datatype display mode
        widget.mode = 'display'
        self.assertEqual(
            widget(),
            '<div class="display-select" id="display-MYSELECT">Two</div>'
        )

        widget.attrs['display_proxy'] = True
        self.checkOutput("""
        <div>
          <div class="display-select" id="display-MYSELECT">Two</div>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="2"/>
        </div>
        """, wrapped_fxml(widget()))

        data = widget.extract(request={'MYSELECT': '1'})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 2)
        self.assertEqual(data.extracted, 1)
        self.assertEqual(data.errors, [])

        self.checkOutput("""
        <div>
          <div class="display-select" id="display-MYSELECT">One</div>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="1"/>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Generic HTML5 Data
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'data': {'foo': 'bar'},
                'vocabulary': [('one', 'One')]
            })
        self.checkOutput("""
        <select class="select" data-foo="bar" id="input-MYSELECT"
                name="MYSELECT">
          <option id="input-MYSELECT-one" selected="selected"
                  value="one">One</option>
        </select>
        """, fxml(widget()))

        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'data': {'foo': 'bar'},
                'vocabulary': [('one', 'One')]
            },
            mode='display')
        self.checkOutput("""
        <div class="display-select" data-foo="bar"
             id="display-MYSELECT">One</div>
        """, fxml(widget()))

        # Persist
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'vocabulary': [('one', 'One')]
            })
        data = widget.extract({'MYSELECT': 'one'})
        model = dict()
        data.persist_writer = write_mapping_writer
        data.write(model)
        self.assertEqual(model, {'MYSELECT': 'one'})

    def test_select_blueprint_single_radio(self):
        # Render single selection as radio inputs
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'vocabulary': vocab,
                'format': 'single',
                'listing_label_position': 'before'
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div id="radio-MYSELECT-wrapper">
            <div id="radio-MYSELECT-one">
              <label for="input-MYSELECT-one">One</label>
              <input checked="checked" class="select" id="input-MYSELECT-one"
                     name="MYSELECT" type="radio" value="one"/>
            </div>
            <div id="radio-MYSELECT-two">
              <label for="input-MYSELECT-two">Two</label>
              <input class="select" id="input-MYSELECT-two" name="MYSELECT"
                     type="radio" value="two"/>
            </div>
            <div id="radio-MYSELECT-three">
              <label for="input-MYSELECT-three">Three</label>
              <input class="select" id="input-MYSELECT-three" name="MYSELECT"
                     type="radio" value="three"/>
            </div>
            <div id="radio-MYSELECT-four">
              <label for="input-MYSELECT-four">Four</label>
              <input class="select" id="input-MYSELECT-four" name="MYSELECT"
                     type="radio" value="four"/>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        # Render single selection as radio inputs, disables all
        widget.attrs['disabled'] = True
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div id="radio-MYSELECT-wrapper">
            <div id="radio-MYSELECT-one">
              <label for="input-MYSELECT-one">One</label>
              <input checked="checked" class="select" disabled="disabled"
                     id="input-MYSELECT-one" name="MYSELECT" type="radio"
                     value="one"/>
            </div>
            <div id="radio-MYSELECT-two">
              <label for="input-MYSELECT-two">Two</label>
              <input class="select" disabled="disabled" id="input-MYSELECT-two"
                     name="MYSELECT" type="radio" value="two"/>
            </div>
            <div id="radio-MYSELECT-three">
              <label for="input-MYSELECT-three">Three</label>
              <input class="select" disabled="disabled"
                     id="input-MYSELECT-three" name="MYSELECT" type="radio"
                     value="three"/>
            </div>
            <div id="radio-MYSELECT-four">
              <label for="input-MYSELECT-four">Four</label>
              <input class="select" disabled="disabled"
                     id="input-MYSELECT-four" name="MYSELECT" type="radio"
                     value="four"/>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        # Render single selection as radio inputs, disables some
        widget.attrs['disabled'] = ['one', 'three']
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div id="radio-MYSELECT-wrapper">
            <div id="radio-MYSELECT-one">
              <label for="input-MYSELECT-one">One</label>
              <input checked="checked" class="select" disabled="disabled"
                     id="input-MYSELECT-one" name="MYSELECT" type="radio"
                     value="one"/>
            </div>
            <div id="radio-MYSELECT-two">
              <label for="input-MYSELECT-two">Two</label>
              <input class="select" id="input-MYSELECT-two" name="MYSELECT"
                     type="radio" value="two"/>
            </div>
            <div id="radio-MYSELECT-three">
              <label for="input-MYSELECT-three">Three</label>
              <input class="select" disabled="disabled"
                     id="input-MYSELECT-three"
                     name="MYSELECT" type="radio" value="three"/>
            </div>
            <div id="radio-MYSELECT-four">
              <label for="input-MYSELECT-four">Four</label>
              <input class="select" id="input-MYSELECT-four" name="MYSELECT"
                     type="radio" value="four"/>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        del widget.attrs['disabled']

        # Radio single valued display mode
        widget.mode = 'display'
        self.assertEqual(
            widget(),
            '<div class="display-select" id="display-MYSELECT">One</div>'
        )

        widget.attrs['display_proxy'] = True
        self.checkOutput("""
        <div>
          <div class="display-select" id="display-MYSELECT">One</div>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="one"/>
        </div>
        """, wrapped_fxml(widget()))

        data = widget.extract(request={'MYSELECT': 'two'})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'one')
        self.assertEqual(data.extracted, 'two')
        self.assertEqual(data.errors, [])

        self.checkOutput("""
        <div>
          <div class="display-select" id="display-MYSELECT">Two</div>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="two"/>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Radio single value selection with uuid datatype set
        vocab = [
            ('3762033b-7118-4bad-89ed-7cb71f5ab6d1', 'One'),
            ('74ef603d-29d0-4016-a003-334719dde835', 'Two'),
            ('b1116392-4a80-496d-86f1-3a2c87e09c59', 'Three'),
            ('e09471dc-625d-463b-be03-438d7089ec13', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value='b1116392-4a80-496d-86f1-3a2c87e09c59',
            props={
                'vocabulary': vocab,
                'datatype': uuid.UUID,
                'format': 'single',
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div id="radio-MYSELECT-wrapper">
            <div id="radio-MYSELECT-3762033b-7118-4bad-89ed-7cb71f5ab6d1">
              <label
                for="input-MYSELECT-3762033b-7118-4bad-89ed-7cb71f5ab6d1"><input
                class="select"
                id="input-MYSELECT-3762033b-7118-4bad-89ed-7cb71f5ab6d1"
                name="MYSELECT" type="radio"
                value="3762033b-7118-4bad-89ed-7cb71f5ab6d1"/>One</label>
            </div>
            <div id="radio-MYSELECT-74ef603d-29d0-4016-a003-334719dde835">
              <label
                for="input-MYSELECT-74ef603d-29d0-4016-a003-334719dde835"><input
                class="select"
                id="input-MYSELECT-74ef603d-29d0-4016-a003-334719dde835"
                name="MYSELECT" type="radio"
                value="74ef603d-29d0-4016-a003-334719dde835"/>Two</label>
            </div>
            <div id="radio-MYSELECT-b1116392-4a80-496d-86f1-3a2c87e09c59">
              <label
                for="input-MYSELECT-b1116392-4a80-496d-86f1-3a2c87e09c59"><input
                checked="checked"
                class="select"
                id="input-MYSELECT-b1116392-4a80-496d-86f1-3a2c87e09c59"
                name="MYSELECT" type="radio"
                value="b1116392-4a80-496d-86f1-3a2c87e09c59"/>Three</label>
            </div>
            <div id="radio-MYSELECT-e09471dc-625d-463b-be03-438d7089ec13">
              <label
                for="input-MYSELECT-e09471dc-625d-463b-be03-438d7089ec13"><input
                class="select"
                id="input-MYSELECT-e09471dc-625d-463b-be03-438d7089ec13"
                name="MYSELECT" type="radio"
                value="e09471dc-625d-463b-be03-438d7089ec13"/>Four</label>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        data = widget.extract({
            'MYSELECT': 'e09471dc-625d-463b-be03-438d7089ec13'
        })
        self.assertEqual(
            data.extracted,
            uuid.UUID('e09471dc-625d-463b-be03-438d7089ec13')
        )

        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div id="radio-MYSELECT-wrapper">
            <div id="radio-MYSELECT-3762033b-7118-4bad-89ed-7cb71f5ab6d1">
              <label
                for="input-MYSELECT-3762033b-7118-4bad-89ed-7cb71f5ab6d1"><input
                class="select"
                id="input-MYSELECT-3762033b-7118-4bad-89ed-7cb71f5ab6d1"
                name="MYSELECT" type="radio"
                value="3762033b-7118-4bad-89ed-7cb71f5ab6d1"/>One</label>
            </div>
            <div id="radio-MYSELECT-74ef603d-29d0-4016-a003-334719dde835">
              <label
                for="input-MYSELECT-74ef603d-29d0-4016-a003-334719dde835"><input
                class="select"
                id="input-MYSELECT-74ef603d-29d0-4016-a003-334719dde835"
                name="MYSELECT" type="radio"
                value="74ef603d-29d0-4016-a003-334719dde835"/>Two</label>
            </div>
            <div id="radio-MYSELECT-b1116392-4a80-496d-86f1-3a2c87e09c59">
              <label
                for="input-MYSELECT-b1116392-4a80-496d-86f1-3a2c87e09c59"><input
                class="select"
                id="input-MYSELECT-b1116392-4a80-496d-86f1-3a2c87e09c59"
                name="MYSELECT" type="radio"
                value="b1116392-4a80-496d-86f1-3a2c87e09c59"/>Three</label>
            </div>
            <div id="radio-MYSELECT-e09471dc-625d-463b-be03-438d7089ec13">
              <label
                for="input-MYSELECT-e09471dc-625d-463b-be03-438d7089ec13"><input
                checked="checked" class="select"
                id="input-MYSELECT-e09471dc-625d-463b-be03-438d7089ec13"
                name="MYSELECT" type="radio"
                value="e09471dc-625d-463b-be03-438d7089ec13"/>Four</label>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Generic HTML5 Data
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'vocabulary': [('one', 'One')],
                'format': 'single',
                'listing_label_position': 'before',
                'data': {'foo': 'bar'}
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div data-foo="bar" id="radio-MYSELECT-wrapper">
            <div id="radio-MYSELECT-one">
              <label for="input-MYSELECT-one">One</label>
              <input checked="checked" class="select" id="input-MYSELECT-one"
                     name="MYSELECT" type="radio" value="one"/>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'vocabulary': [('one', 'One')],
                'format': 'single',
                'listing_label_position': 'before',
                'data': {'foo': 'bar'}
            },
            mode='display')
        self.checkOutput("""
        <div>
          <div class="display-select" data-foo="bar"
               id="display-MYSELECT">One</div>
        </div>
        """, wrapped_fxml(widget()))

    def test_select_blueprint_multi(self):
        # Empty multi valued
        widget = factory(
            'select',
            name='EMPTYSELECT',
            value=UNSET,
            props={
                'multivalued': True,
                'vocabulary': []
            })
        self.checkOutput("""
        <div>
          <input id="exists-EMPTYSELECT" name="EMPTYSELECT-exists"
                 type="hidden" value="exists"/>
          <select class="select" id="input-EMPTYSELECT" multiple="multiple"
                  name="EMPTYSELECT"> </select>
        </div>
        """, wrapped_fxml(widget()))

        # Default multi valued
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value=['one', 'two'],
            props={
                'multivalued': True,
                'vocabulary': vocab
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <select class="select" id="input-MYSELECT" multiple="multiple"
                  name="MYSELECT">
            <option id="input-MYSELECT-one" selected="selected"
                    value="one">One</option>
            <option id="input-MYSELECT-two" selected="selected"
                    value="two">Two</option>
            <option id="input-MYSELECT-three" value="three">Three</option>
            <option id="input-MYSELECT-four" value="four">Four</option>
          </select>
        </div>
        """, wrapped_fxml(widget()))

        # Extract multi valued selection and render widget with extracted data
        data = widget.extract(request={'MYSELECT': ['one', 'four']})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, ['one', 'two'])
        self.assertEqual(data.extracted, ['one', 'four'])
        self.assertEqual(data.errors, [])

        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <select class="select" id="input-MYSELECT" multiple="multiple"
                  name="MYSELECT">
            <option id="input-MYSELECT-one" selected="selected"
                    value="one">One</option>
            <option id="input-MYSELECT-two" value="two">Two</option>
            <option id="input-MYSELECT-three" value="three">Three</option>
            <option id="input-MYSELECT-four" selected="selected"
                    value="four">Four</option>
          </select>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Multi selection display mode
        widget.mode = 'display'
        self.checkOutput("""
        <ul class="display-select" id="display-MYSELECT">
          <li>One</li>
          <li>Two</li>
        </ul>
        """, fxml(widget()))

        # Multi selection display mode with display proxy
        widget.attrs['display_proxy'] = True
        self.checkOutput("""
        <div>
          <ul class="display-select" id="display-MYSELECT">
            <li>One</li>
            <li>Two</li>
          </ul>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="one"/>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="two"/>
        </div>
        """, wrapped_fxml(widget()))

        # Multi selection display mode with display proxy and extracted data
        data = widget.extract(request={'MYSELECT': ['one']})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, ['one', 'two'])
        self.assertEqual(data.extracted, ['one'])
        self.assertEqual(data.errors, [])

        self.checkOutput("""
        <div>
          <ul class="display-select" id="display-MYSELECT">
            <li>One</li>
          </ul>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="one"/>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Multi selection display with empty values list
        widget = factory(
            'select',
            name='MYSELECT',
            value=[],
            props={
                'vocabulary': [],
                'multivalued': True
            },
            mode='display')
        self.checkOutput("""
        <div>
          <div class="display-select" id="display-MYSELECT"/>
        </div>
        """, wrapped_fxml(widget()))

        # Multi selection display with missing term in vocab
        widget = factory(
            'select',
            name='MYSELECT',
            value=['one', 'two'],
            props={
                'multivalued': True,
                'vocabulary': [('two', 'Two')]
            },
            mode='display')
        self.checkOutput("""
        <ul class="display-select" id="display-MYSELECT">
          <li>one</li>
          <li>Two</li>
        </ul>
        """, fxml(widget()))

        # Multiple values on single valued selection fails
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value=['one', 'two'],
            props={
                'vocabulary': vocab
            })
        with self.assertRaises(ValueError) as arc:
            widget()
        self.assertEqual(
            str(arc.exception),
            'Multiple values for single selection.'
        )

        # Multi value selection with float datatype set
        vocab = [
            (1.0, 'One'),
            (2.0, 'Two'),
            (3.0, 'Three'),
            (4.0, 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value=[1.0, 2.0],
            props={
                'datatype': float,
                'multivalued': True,
                'vocabulary': vocab,
                'emptyvalue': []
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <select class="select" id="input-MYSELECT" multiple="multiple"
                  name="MYSELECT">
            <option id="input-MYSELECT-1.0" selected="selected"
                    value="1.0">One</option>
            <option id="input-MYSELECT-2.0" selected="selected"
                    value="2.0">Two</option>
            <option id="input-MYSELECT-3.0" value="3.0">Three</option>
            <option id="input-MYSELECT-4.0" value="4.0">Four</option>
          </select>
        </div>
        """, wrapped_fxml(widget()))

        request = {
            'MYSELECT': ['2.0', '3.0']
        }
        data = widget.extract(request=request)
        self.assertEqual(data.extracted, [2.0, 3.0])

        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <select class="select" id="input-MYSELECT" multiple="multiple"
                  name="MYSELECT">
            <option id="input-MYSELECT-1.0" value="1.0">One</option>
            <option id="input-MYSELECT-2.0" selected="selected"
                    value="2.0">Two</option>
            <option id="input-MYSELECT-3.0" selected="selected"
                    value="3.0">Three</option>
            <option id="input-MYSELECT-4.0" value="4.0">Four</option>
          </select>
        </div>
        """, wrapped_fxml(widget(data=data)))

        request = {
            'MYSELECT': '4.0'
        }
        data = widget.extract(request=request)
        self.assertEqual(data.extracted, [4.0])

        request = {
            'MYSELECT': ''
        }
        data = widget.extract(request=request)
        self.assertEqual(data.extracted, [])

        # Generic HTML5 Data
        vocab = [
            ('one', 'One'),
            ('two', 'Two')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value=['one', 'two'],
            props={
                'multivalued': True,
                'data': {'foo': 'bar'},
                'vocabulary': vocab
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
            value="exists"/>
          <select class="select" data-foo="bar" id="input-MYSELECT"
                  multiple="multiple" name="MYSELECT">
            <option id="input-MYSELECT-one" selected="selected"
                    value="one">One</option>
            <option id="input-MYSELECT-two" selected="selected"
                    value="two">Two</option>
          </select>
        </div>
        """, wrapped_fxml(widget()))

        widget.mode = 'display'
        self.checkOutput("""
        <ul class="display-select" data-foo="bar" id="display-MYSELECT">
          <li>One</li>
          <li>Two</li>
        </ul>
        """, fxml(widget()))

        # Persist
        widget = factory(
            'select',
            name='MYSELECT',
            value=['one', 'two'],
            props={
                'multivalued': True,
                'vocabulary': vocab
            })
        data = widget.extract({'MYSELECT': ['one', 'two', 'three']})
        model = dict()
        data.persist_writer = write_mapping_writer
        data.write(model)
        self.assertEqual(model, {'MYSELECT': ['one', 'two', 'three']})

    def test_select_blueprint_multi_checkboxes(self):
        # Render multi selection as checkboxes
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'multivalued': True,
                'vocabulary': vocab,
                'format': 'single'
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div id="checkbox-MYSELECT-wrapper">
            <div id="checkbox-MYSELECT-one">
              <label for="input-MYSELECT-one"><input checked="checked"
                     class="select" id="input-MYSELECT-one" name="MYSELECT"
                     type="checkbox" value="one"/>One</label>
            </div>
            <div id="checkbox-MYSELECT-two">
              <label for="input-MYSELECT-two"><input class="select"
                     id="input-MYSELECT-two" name="MYSELECT" type="checkbox"
                     value="two"/>Two</label>
            </div>
            <div id="checkbox-MYSELECT-three">
              <label for="input-MYSELECT-three"><input class="select"
                     id="input-MYSELECT-three" name="MYSELECT" type="checkbox"
                     value="three"/>Three</label>
            </div>
            <div id="checkbox-MYSELECT-four">
              <label for="input-MYSELECT-four"><input class="select"
                     id="input-MYSELECT-four" name="MYSELECT" type="checkbox"
                     value="four"/>Four</label>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        # Checkbox multi selection display mode. Note, other as above, preset
        # value for multivalued widget is set as string, which is treaten as
        # one item selected and covered with the below tests
        widget.mode = 'display'
        self.checkOutput("""
        <ul class="display-select" id="display-MYSELECT">
          <li>One</li>
        </ul>
        """, fxml(widget()))

        # Checkbox multi selection display mode with display proxy
        widget.attrs['display_proxy'] = True
        self.checkOutput("""
        <div>
          <ul class="display-select" id="display-MYSELECT">
            <li>One</li>
          </ul>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="one"/>
        </div>
        """, wrapped_fxml(widget()))

        # Checkbox multi selection display mode with display proxy and
        # extracted data
        data = widget.extract(request={'MYSELECT': ['two']})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'one')
        self.assertEqual(data.extracted, ['two'])
        self.assertEqual(data.errors, [])

        self.checkOutput("""
        <div>
          <ul class="display-select" id="display-MYSELECT">
            <li>Two</li>
          </ul>
          <input class="select" id="input-MYSELECT" name="MYSELECT"
                 type="hidden" value="two"/>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Generic HTML5 Data
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'multivalued': True,
                'data': {'foo': 'bar'},
                'vocabulary': [('one', 'One')],
                'format': 'single'
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div data-foo="bar" id="checkbox-MYSELECT-wrapper">
            <div id="checkbox-MYSELECT-one">
              <label for="input-MYSELECT-one"><input checked="checked"
                     class="select" id="input-MYSELECT-one" name="MYSELECT"
                     type="checkbox" value="one"/>One</label>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        widget.mode = 'display'
        self.checkOutput("""
        <ul class="display-select" data-foo="bar" id="display-MYSELECT">
          <li>One</li>
        </ul>
        """, fxml(widget()))

    def test_select_blueprint_misc(self):
        # Using 'ul' instead of 'div' for rendering radio or checkbox
        # selections
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'multivalued': True,
                'vocabulary': vocab,
                'format': 'single',
                'listing_tag': 'ul'
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <ul id="checkbox-MYSELECT-wrapper">
            <li id="checkbox-MYSELECT-one">
              <label for="input-MYSELECT-one"><input checked="checked"
                     class="select" id="input-MYSELECT-one" name="MYSELECT"
                     type="checkbox" value="one"/>One</label>
            </li>
            <li id="checkbox-MYSELECT-two">
              <label for="input-MYSELECT-two"><input class="select"
                     id="input-MYSELECT-two" name="MYSELECT" type="checkbox"
                     value="two"/>Two</label>
            </li>
            <li id="checkbox-MYSELECT-three">
              <label for="input-MYSELECT-three"><input class="select"
                     id="input-MYSELECT-three" name="MYSELECT" type="checkbox"
                     value="three"/>Three</label>
            </li>
            <li id="checkbox-MYSELECT-four">
              <label for="input-MYSELECT-four"><input class="select"
                     id="input-MYSELECT-four" name="MYSELECT" type="checkbox"
                     value="four"/>Four</label>
            </li>
          </ul>
        </div>
        """, wrapped_fxml(widget()))

        # Render single format selection with label after input
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'multivalued': True,
                'vocabulary': [
                    ('one', 'One'),
                    ('two', 'Two'),
                ],
                'format': 'single',
                'listing_tag': 'ul',
                'listing_label_position': 'after'
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <ul id="checkbox-MYSELECT-wrapper">
            <li id="checkbox-MYSELECT-one">
              <input checked="checked" class="select" id="input-MYSELECT-one"
                     name="MYSELECT" type="checkbox" value="one"/>
              <label for="input-MYSELECT-one">One</label>
            </li>
            <li id="checkbox-MYSELECT-two">
              <input class="select" id="input-MYSELECT-two" name="MYSELECT"
                     type="checkbox" value="two"/>
              <label for="input-MYSELECT-two">Two</label>
            </li>
          </ul>
        </div>
        """, wrapped_fxml(widget()))

        # Render single format selection with input inside label before
        # checkbox
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'multivalued': True,
                'vocabulary': [
                    ('one', 'One'),
                    ('two', 'Two'),
                ],
                'format': 'single',
                'listing_tag': 'ul',
                'listing_label_position': 'inner-before'
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <ul id="checkbox-MYSELECT-wrapper">
            <li id="checkbox-MYSELECT-one">
              <label for="input-MYSELECT-one">One<input checked="checked"
                     class="select" id="input-MYSELECT-one" name="MYSELECT"
                     type="checkbox" value="one"/></label>
            </li>
            <li id="checkbox-MYSELECT-two">
              <label for="input-MYSELECT-two">Two<input class="select"
                     id="input-MYSELECT-two" name="MYSELECT" type="checkbox"
                     value="two"/></label>
            </li>
          </ul>
        </div>
        """, wrapped_fxml(widget()))

        # Check BBB 'inner' for 'listing_label_position' which behaves like
        # 'inner-after'
        widget = factory(
            'select',
            name='MYSELECT',
            value='one',
            props={
                'vocabulary': [('one', 'One')],
                'format': 'single',
                'listing_label_position': 'inner'
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <div id="radio-MYSELECT-wrapper">
            <div id="radio-MYSELECT-one">
              <label for="input-MYSELECT-one"><input checked="checked"
                     class="select" id="input-MYSELECT-one" name="MYSELECT"
                     type="radio" value="one"/>One</label>
            </div>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        # Check selection required
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'required': 'Selection required',
                'vocabulary': vocab
            })
        self.checkOutput("""
        <select class="select" id="input-MYSELECT" name="MYSELECT"
                required="required">
          <option id="input-MYSELECT-one" value="one">One</option>
          <option id="input-MYSELECT-two" value="two">Two</option>
          <option id="input-MYSELECT-three" value="three">Three</option>
          <option id="input-MYSELECT-four" value="four">Four</option>
        </select>
        """, fxml(widget()))

        data = widget.extract(request={'MYSELECT': ''})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, '')
        self.assertEqual(data.errors, [ExtractionError('Selection required')])

        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'required': 'Selection required',
                'multivalued': True,
                'vocabulary': vocab
            })
        self.checkOutput("""
        <div>
          <input id="exists-MYSELECT" name="MYSELECT-exists" type="hidden"
                 value="exists"/>
          <select class="select" id="input-MYSELECT" multiple="multiple"
                  name="MYSELECT" required="required">
            <option id="input-MYSELECT-one" value="one">One</option>
            <option id="input-MYSELECT-two" value="two">Two</option>
            <option id="input-MYSELECT-three" value="three">Three</option>
            <option id="input-MYSELECT-four" value="four">Four</option>
          </select>
        </div>
        """, wrapped_fxml(widget()))

        data = widget.extract(request={'MYSELECT-exists': 'exists'})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, [])
        self.assertEqual(data.errors, [ExtractionError('Selection required')])

        # Check selection required with datatype set
        vocab = [
            (1, 'One'),
            (2, 'Two'),
            (3, 'Three'),
            (4, 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'required': 'Selection required',
                'multivalued': True,
                'vocabulary': vocab,
                'datatype': int,
            })
        data = widget.extract(request={'MYSELECT-exists': 'exists'})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, [])
        self.assertEqual(data.errors, [ExtractionError('Selection required')])

        data = widget.extract(request={'MYSELECT': ['1', '2']})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, [1, 2])
        self.assertEqual(data.errors, [])

        # Single selection extraction without value
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'vocabulary': [
                    ('one', 'One'),
                    ('two', 'Two')
                ]
            })
        request = {
            'MYSELECT': 'one',
            'MYSELECT-exists': True,
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'one')
        self.assertEqual(data.errors, [])

        # Single selection extraction with value
        widget = factory(
            'select',
            name='MYSELECT',
            value='two',
            props={
                'vocabulary': [
                    ('one', 'One'),
                    ('two', 'Two')
                ]
            })
        request = {
            'MYSELECT': 'one',
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'two')
        self.assertEqual(data.extracted, 'one')
        self.assertEqual(data.errors, [])

        # Single selection extraction disabled (means browser does not post the
        # value) with value
        widget.attrs['disabled'] = True
        data = widget.extract({'MYSELECT-exists': True})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'two')
        self.assertEqual(data.extracted, 'two')
        self.assertEqual(data.errors, [])

        # Disabled can be also the value itself
        widget.attrs['disabled'] = 'two'
        data = widget.extract({'MYSELECT-exists': True})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'two')
        self.assertEqual(data.extracted, 'two')
        self.assertEqual(data.errors, [])

        # Single selection extraction required
        widget = factory(
            'select',
            name='MYSELECT',
            value='two',
            props={
                'required': True,
                'vocabulary': [
                    ('one', 'One'),
                    ('two', 'Two')
                ]
            })
        request = {
            'MYSELECT': '',
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'two')
        self.assertEqual(data.extracted, '')
        self.assertEqual(
            data.errors,
            [ExtractionError('Mandatory field was empty')]
        )

        # A disabled and required returns value itself
        widget.attrs['disabled'] = True
        data = widget.extract({'MYSELECT-exists': True})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'two')
        self.assertEqual(data.extracted, 'two')
        self.assertEqual(data.errors, [])

        # Multiple selection extraction without value
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'multivalued': True,
                'vocabulary': [
                    ('one', 'One'),
                    ('two', 'Two')
                ]
            })
        request = {
            'MYSELECT': ['one', 'two'],
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, ['one', 'two'])
        self.assertEqual(data.errors, [])

        # Multiple selection extraction with value
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value='three',
            props={
                'multivalued': True,
                'vocabulary': vocab
            })
        request = {
            'MYSELECT': 'one',
            'MYSELECT-exists': True,
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'three')
        self.assertEqual(data.extracted, ['one'])
        self.assertEqual(data.errors, [])

        # Multiselection, completly disabled
        widget.attrs['disabled'] = True
        data = widget.extract({'MYSELECT-exists': True})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, 'three')
        self.assertEqual(data.extracted, ['three'])
        self.assertEqual(data.errors, [])

        # Multiselection, partly disabled, empty request
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value=['one', 'three'],
            props={
                'multivalued': True,
                'disabled': ['two', 'three'],
                'vocabulary': vocab
            })
        data = widget.extract({})
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, ['one', 'three'])
        self.assertEqual(data.extracted, UNSET)
        self.assertEqual(data.errors, [])

        # Multiselection, partly disabled, non-empty request
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three'),
            ('four', 'Four'),
            ('five', 'Five')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            value=['one', 'two', 'four'],
            props={
                'multivalued': True,
                'disabled': ['two', 'three', 'four', 'five'],
                'vocabulary': vocab,
                'datatype': UNICODE_TYPE,
            })
        request = {
            'MYSELECT': ['one', 'two', 'five'],
            'MYSELECT-exists': True,
        }

        # Explanation:
        #
        # * one is a simple value as usal,
        # * two is disabled and in value, so it should be kept in.
        # * three is disabled and not in value, so it should kept out,
        # * four is disabled and in value, but someone removed it in the
        #   request, it should get recovered,
        # * five is disabled and not in value, but someone put it in the
        #   request. it should get removed.

        # Check extraction

        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, ['one', 'two', 'four'])
        self.assertEqual(data.extracted, [u'one', u'two', u'four'])
        self.assertEqual(data.errors, [])

        # Single selection radio extraction
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three')
        ]
        widget = factory(
            'select',
            'MYSELECT',
            props={
                'format': 'single',
                'vocabulary': vocab
            })

        # No exists marker in request. Extracts to UNSET
        request = {}
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, UNSET)
        self.assertEqual(data.errors, [])

        # Exists marker in request. Extracts to empty string
        request = {
            'MYSELECT-exists': '1',
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, '')
        self.assertEqual(data.errors, [])

        # Select value
        request = {
            'MYSELECT-exists': '1',
            'MYSELECT': 'one',
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'one')
        self.assertEqual(data.errors, [])

        # Multi selection radio extraction
        vocab = [
            ('one', 'One'),
            ('two', 'Two'),
            ('three', 'Three')
        ]
        widget = factory(
            'select',
            name='MYSELECT',
            props={
                'multivalued': True,
                'format': 'single',
                'vocabulary': vocab
            })

        # No exists marker in request. Extracts to UNSET
        request = {}
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, UNSET)
        self.assertEqual(data.errors, [])

        # Exists marker in request. Extracts to empty list
        request = {
            'MYSELECT-exists': '1',
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, [])
        self.assertEqual(data.errors, [])

        # Select values
        request = {
            'MYSELECT-exists': '1',
            'MYSELECT': ['one', 'two'],
        }
        data = widget.extract(request)
        self.assertEqual(data.name, 'MYSELECT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, ['one', 'two'])
        self.assertEqual(data.errors, [])



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import factory
from yafowil.common import empty_display_renderer
from yafowil.common import generic_extractor
from yafowil.datatypes import generic_datatype_extractor
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops


###############################################################################
# proxy
###############################################################################

@managedprops(*css_managed_props)
def input_proxy_renderer(widget, data):
    """Render hidden input ignoring ``widget.dottedpath``, just using widget
    name.
    """
    tag = data.tag
    value = data.value
    if data.request is not UNSET and data.request.get(widget.__name__):
        value = data.request.get(widget.__name__)
    input_attrs = {
        'type': 'hidden',
        'value': value,
        'name_': widget.__name__,
        'id': cssid(widget, 'input'),
        'class_': cssclasses(widget, data),
    }
    return tag('input', **input_attrs)


factory.register(
    'proxy',
    extractors=[
        generic_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor
    ],
    edit_renderers=[input_proxy_renderer],
    display_renderers=[empty_display_renderer]
)

factory.doc['blueprint']['proxy'] = """\
Bypass arguments out of form namespace using a hidden field.
"""

factory.defaults['proxy.class'] = None



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.common import input_generic_renderer
from yafowil.datatypes import generic_emptyvalue_extractor


###############################################################################
# search
###############################################################################

factory.register(
    'search',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['search'] = """\
Search blueprint (HTML5).
"""

factory.defaults['search.type'] = 'search'

factory.defaults['search.default'] = ''

factory.defaults['search.required_class'] = 'required'

factory.defaults['search.class'] = 'search'



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops


###############################################################################
# tag
###############################################################################

@managedprops('tag', 'text', 'data', *css_managed_props)
def tag_renderer(widget, data):
    """Renderer for HTML tags.
    """
    attrs = {
        'id': cssid(widget, 'tag'),
        'class_': cssclasses(widget, data),
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    tag = attr_value('tag', widget, data)
    text = attr_value('text', widget, data)
    return data.tag(tag, text, **attrs)


factory.register(
    'tag',
    edit_renderers=[tag_renderer],
    display_renderers=[tag_renderer]
)

factory.doc['blueprint']['tag'] = """\
Render HTML tags with text. Useful for rendering headings etc.
"""

factory.doc['props']['tag.tag'] = """\
HTML tag name.
"""

factory.doc['props']['tag.text'] = """\
Tag contents.
"""



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase


class TestNumber(YafowilTestCase):

    def test_number_blueprint(self):
        # Display renderer
        widget = factory(
            'number',
            name='NUMBER',
            value=3,
            mode='display')
        self.assertEqual(widget(), (
            '<div class="display-number" id="display-NUMBER">3</div>'
        ))

        # Render number input
        widget = factory(
            'number',
            name='NUMBER',
            value=lambda w, d: 3)
        self.assertEqual(widget(), (
            '<input class="number" id="input-NUMBER" name="NUMBER" '
            'type="number" value="3" />'
        ))

        # Extract unset
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        self.assertEqual(data.name, 'NUMBER')
        self.assertEqual(data.value, 3)
        self.assertEqual(data.extracted, UNSET)
        self.assertEqual(data.errors, [])

        # Extract not required and empty
        data = widget.extract({'NUMBER': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        # Extract invalid floating point input
        data = widget.extract({'NUMBER': 'abc'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input is not a valid floating point number.')]
        )

        # Extract valid floating point input
        data = widget.extract({'NUMBER': '10'})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 10.)

        data = widget.extract({'NUMBER': '10.0'})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 10.)

        data = widget.extract({'NUMBER': '10,0'})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 10.)

        # Instanciate with invalid datatype
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'datatype': 'invalid'
            })
        with self.assertRaises(ValueError) as arc:
            widget.extract({'NUMBER': '10.0'})
        self.assertEqual(str(arc.exception), 'Datatype not allowed: "invalid"')

        # Extract invalid integer input
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'datatype': int
            })
        data = widget.extract({'NUMBER': '10.0'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input is not a valid integer.')]
        )

        # Extract with min value set
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'min': 10
            })
        data = widget.extract({'NUMBER': '9'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Value has to be at minimum 10.')]
        )

        data = widget.extract({'NUMBER': '10'})
        self.assertEqual(data.errors, [])

        data = widget.extract({'NUMBER': '11'})
        self.assertEqual(data.errors, [])

        # Extract min value 0
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'min': 0
            })
        data = widget.extract({'NUMBER': '-1'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Value has to be at minimum 0.')]
        )

        data = widget.extract({'NUMBER': '0'})
        self.assertEqual(data.errors, [])

        # Extract with max value set
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'max': lambda w, d: 10
            })
        data = widget.extract({'NUMBER': '9'})
        self.assertEqual(data.errors, [])

        data = widget.extract({'NUMBER': '10'})
        self.assertEqual(data.errors, [])

        data = widget.extract({'NUMBER': '11'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Value has to be at maximum 10.')]
        )

        # Extract max value 0
        widget = factory(
            'number',
            name='NUMBER',
            props={'max': 0}
        )
        data = widget.extract({'NUMBER': '1'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Value has to be at maximum 0.')]
        )

        data = widget.extract({'NUMBER': '0'})
        self.assertEqual(data.errors, [])

        # Extract with step set
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'step': 2
            })
        data = widget.extract({'NUMBER': '9'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Value 9.0 has to be in stepping of 2')]
        )

        data = widget.extract({'NUMBER': '6'})
        self.assertEqual(data.errors, [])

        # Extract with step and min value set
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'step': 2,
                'min': 3
            })
        data = widget.extract({'NUMBER': '7'})
        self.assertEqual(data.errors, [])

        data = widget.extract({'NUMBER': '6'})
        self.assertEqual(
            data.errors,
            [ExtractionError(
                'Value 6.0 has to be in stepping of 2 based on a floor '
                'value of 3'
            )]
        )

        # Extract 0 value
        widget = factory(
            'number',
            name='NUMBER')
        data = widget.extract({'NUMBER': '0'})
        self.assertEqual(data.extracted, 0.0)

        widget = factory(
            'number',
            name='NUMBER',
            props={
                'datatype': int
            })
        data = widget.extract({'NUMBER': '0'})
        self.assertEqual(data.extracted, 0)

        # Persist
        widget = factory(
            'number',
            name='NUMBER',
            props={
                'datatype': int
            })
        data = widget.extract({'NUMBER': '0'})
        model = dict()
        data.persist_writer = write_mapping_writer
        data.write(model)
        self.assertEqual(model, {'NUMBER': 0})



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.tests import YafowilTestCase


class TestSearch(YafowilTestCase):

    def test_search_blueprint(self):
        # Render search input field
        widget = factory(
            'search',
            name='SEARCH')
        self.assertEqual(widget(), (
            '<input class="search" id="input-SEARCH" name="SEARCH" '
            'type="search" value="" />'
        ))

        # Extract not required and empty
        data = widget.extract({'SEARCH': ''})
        self.assertEqual(data.errors, [])

        # Extract required empty
        widget.attrs['required'] = True
        data = widget.extract({'SEARCH': ''})
        self.assertEqual(data.name, 'SEARCH')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, '')
        self.assertEqual(
            data.errors,
            [ExtractionError('Mandatory field was empty')]
        )

        del widget.attrs['required']

        # Emptyvalue
        widget.attrs['emptyvalue'] = 'defaultsearch'
        data = widget.extract(request={'SEARCH': ''})
        self.assertEqual(data.name, 'SEARCH')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'defaultsearch')
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'SEARCH': 'searchstr'})
        self.assertEqual(data.name, 'SEARCH')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'searchstr')
        self.assertEqual(data.errors, [])



from yafowil.base import factory
from yafowil.tests import YafowilTestCase


class TestButton(YafowilTestCase):

    def test_submit_blueprint(self):
        # Render submit button
        widget = factory(
            'submit',
            name='SAVE',
            props={
                'action': True,
                'label': 'Action name',
            })
        self.assertEqual(widget(), (
            '<input id="input-SAVE" name="action.SAVE" type="submit" '
            'value="Action name" />'
        ))

        # If expression is or evaluates to False, skip rendering
        widget = factory(
            'submit',
            name='SAVE',
            props={
                'action': True,
                'label': 'Action name',
                'expression': False,
            })
        self.assertEqual(widget(), '')

        widget = factory(
            'submit',
            name='SAVE',
            props={
                'action': True,
                'label': 'Action name',
                'expression': lambda w, d: False,
            })
        self.assertEqual(widget(), '')

        # Generic HTML5 Data
        widget = factory(
            'submit',
            name='SAVE',
            props={
                'action': True,
                'label': 'Action name',
                'data': {'foo': 'bar'},
            })
        self.assertEqual(widget(), (
            '<input data-foo=\'bar\' id="input-SAVE" name="action.SAVE" '
            'type="submit" value="Action name" />'
        ))

    def test_button_blueprint(self):
        # Render button element
        widget = factory(
            'button',
            name='SAVE',
            props={
                'action': True,
                'text': 'Button text',
            })
        self.assertEqual(
            widget(),
            '<button id="input-SAVE" name="action.SAVE" type="submit">'
            'Button text</button>'
        )

        # If expression is or evaluates to False, skip rendering
        widget = factory(
            'button',
            name='SAVE',
            props={
                'action': True,
                'text': 'Button text',
                'expression': False,
            })
        self.assertEqual(widget(), '')

        widget = factory(
            'button',
            name='SAVE',
            props={
                'action': True,
                'text': 'Button text',
                'expression': lambda w, d: False,
            })
        self.assertEqual(widget(), '')

        # Generic HTML5 Data
        widget = factory(
            'button',
            name='SAVE',
            props={
                'action': True,
                'text': 'Button text',
                'data': {'foo': 'bar'},
            })
        self.assertEqual(
            widget(),
            """<button data-foo='bar' id="input-SAVE" name="action.SAVE" """
            """type="submit">Button text</button>"""
        )

        # Button specific attrs
        widget = factory(
            'button',
            name='SAVE',
            props={
                'action': True,
                'text': 'Button text',
                'form': 'my-parent-for',
                'formaction': 'alternative-action',
                'formenctype': 'text/plain',
                'formmethod': 'post',
                'formnovalidate': "1",
                'formtarget': "_blank",
            })
        self.assertEqual(
            widget(),
            '<button form="my-parent-for" formaction="alternative-action" '
            'formenctype="text/plain" formmethod="post" formnovalidate="1" '
            'formtarget="_blank" id="input-SAVE" name="action.SAVE" '
            'type="submit">Button text</button>'
        )

        # Custom button name if action not defined
        widget = factory(
            'button',
            name='BUTTON',
            props={
                'text': 'Button text',
                'action': False,
                'name': 'custom_button_action'
            })
        self.assertEqual(
            widget(),
            '<button id="input-BUTTON" name="custom_button_action" '
            'type="submit">Button text</button>'
        )



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase


class TestUrl(YafowilTestCase):

    def test_url_blueprint(self):
        # Render URL input field
        widget = factory(
            'url',
            name='URL')
        self.assertEqual(widget(), (
            '<input class="url" id="input-URL" name="URL" '
            'type="url" value="" />'
        ))

        # Extract not required and empty
        data = widget.extract({'URL': ''})
        self.assertEqual(data.errors, [])

        # Extract invalid URL input
        data = widget.extract({'URL': 'htt:/bla'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input not a valid web address.')]
        )

        data = widget.extract({'URL': 'invalid'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input not a valid web address.')]
        )

        # Extract value URL input
        data = widget.extract({
            'URL': 'http://www.foo.bar.com:8080/bla#fasel?blubber=bla&bla=fasel'  # noqa
        })
        self.assertEqual(data.errors, [])

        # Emptyvalue
        widget = factory(
            'url',
            name='URL',
            props={
                'emptyvalue': 'http://www.example.com',
            })
        data = widget.extract(request={'URL': ''})
        self.assertEqual(data.name, 'URL')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'http://www.example.com')
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'URL': 'http://www.example.org'})
        self.assertEqual(data.name, 'URL')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'http://www.example.org')
        self.assertEqual(data.errors, [])

        # Persist
        widget = factory(
            'url',
            name='URL')
        data = widget.extract({'URL': 'http://www.example.org'})
        model = dict()
        data.persist_writer = write_mapping_writer
        data.write(model)
        self.assertEqual(model, {'URL': 'http://www.example.org'})
